import dataclasses as dc
import typing as ta

from ..... import check
from ..... import lang
from ....debug import DEBUG
from ....specs import CoerceFn
from ....specs import FieldSpec
from ....specs import ReprFn
from ....specs import ValidateFn
from ....tools.modifiers import field_modifier
from ...utils import chain_mapping_proxy


##


class _ExtraFieldParamsMetadata(lang.Marker):
    pass


# The following function duplicates non-stdlib kwargs/defaults from the field constructor, but does so only for
# autocomplete - the underlying impl blindly forwards given kwargs to allow distinction between `None` and absent.


def extra_field_params(
        *,

        coerce: bool | CoerceFn | None = None,
        validate: ValidateFn | None = None,
        check_type: bool | type | tuple[type | None, ...] | None = None,
        override: bool | None = None,
        repr_fn: ReprFn | None = None,
        repr_priority: int | None = None,
):
    return {}


def _extra_field_params(**kwargs: ta.Any) -> ta.Mapping[ta.Any, ta.Any]:
    return {_ExtraFieldParamsMetadata: kwargs}


globals()['extra_field_params'] = _extra_field_params


##


def get_field_spec(f: dc.Field) -> FieldSpec | None:
    try:
        fs = f.metadata[FieldSpec]
    except KeyError:
        return None

    return check.isinstance(fs, FieldSpec)


##


def set_field_metadata(
        f: dc.Field,
        metadata: ta.Mapping[ta.Any, ta.Any],
) -> dc.Field:
    if DEBUG:
        check.isinstance(f, dc.Field)

    md: ta.Any = f.metadata

    mdu: dict = {}
    for k, v in metadata.items():
        if md is None or md.get(k) != v:
            mdu[k] = v  # noqa
    if not mdu:
        return f

    if md is None:
        ms = [mdu]
    else:
        ms = [mdu, md]

    f.metadata = chain_mapping_proxy(*ms)
    return f


#


def set_field_spec_metadata(
        f: dc.Field,
        fs: FieldSpec,
) -> None:
    set_field_metadata(
        f,
        {
            FieldSpec: fs,
            _ExtraFieldParamsMetadata: {},
        },
    )


##


def update_extra_field_params(
        f: dc.Field,
        *,
        unless_non_default: bool = False,
        **kwargs: ta.Any,
) -> dc.Field:
    fe = f.metadata.get(_ExtraFieldParamsMetadata, {})
    return set_field_metadata(f, {
        _ExtraFieldParamsMetadata: {
            **(fe if not unless_non_default else {}),
            **kwargs,
            **(fe if unless_non_default else {}),
        },
    })


def with_extra_field_params(**kwargs: ta.Any) -> field_modifier:
    @field_modifier
    def inner(f: dc.Field) -> dc.Field:
        return update_extra_field_params(f, **kwargs)
    return inner



# ruff: noqa: UP006
import abc
import typing as ta


T = ta.TypeVar('T')


##


_ABSTRACT_METHODS_ATTR = '__abstractmethods__'
_IS_ABSTRACT_METHOD_ATTR = '__isabstractmethod__'


def is_abstract_method(obj: ta.Any) -> bool:
    return bool(getattr(obj, _IS_ABSTRACT_METHOD_ATTR, False))


def compute_abstract_methods(cls: type) -> ta.FrozenSet[str]:
    # ~> https://github.com/python/cpython/blob/f3476c6507381ca860eec0989f53647b13517423/Modules/_abc.c#L358

    # Stage 1: direct abstract methods

    abstracts = {
        a
        # Get items as a list to avoid mutation issues during iteration
        for a, v in list(cls.__dict__.items())
        if is_abstract_method(v)
    }

    # Stage 2: inherited abstract methods

    for base in cls.__bases__:
        # Get __abstractmethods__ from base if it exists
        if (base_abstracts := getattr(base, _ABSTRACT_METHODS_ATTR, None)) is None:
            continue

        # Iterate over abstract methods in base
        for key in base_abstracts:
            # Check if this class has an attribute with this name
            try:
                value = getattr(cls, key)
            except AttributeError:
                # Attribute not found in this class, skip
                continue

            # Check if it's still abstract
            if is_abstract_method(value):
                abstracts.add(key)

    return frozenset(abstracts)


def update_abstracts(cls: ta.Type[T], *, force: bool = False) -> ta.Type[T]:
    if not force and not hasattr(cls, _ABSTRACT_METHODS_ATTR):
        # Per stdlib: We check for __abstractmethods__ here because cls might by a C implementation or a python
        # implementation (especially during testing), and we want to handle both cases.
        return cls

    abstracts = compute_abstract_methods(cls)
    setattr(cls, _ABSTRACT_METHODS_ATTR, abstracts)
    return cls


#


class AbstractTypeError(TypeError):
    pass


_FORCE_ABSTRACT_ATTR = '__forceabstract__'


class Abstract:
    """
    Different from, but interoperable with, abc.ABC / abc.ABCMeta:

     - This raises AbstractTypeError during class creation, not instance instantiation - unless Abstract or abc.ABC are
       explicitly present in the class's direct bases.
     - This will forbid instantiation of classes with Abstract in their direct bases even if there are no
       abstractmethods left on the class.
     - This is a mixin, not a metaclass.
     - As it is not an ABCMeta, this does not support virtual base classes. As a result, operations like `isinstance`
       and `issubclass` are ~7x faster.
     - It additionally enforces a base class order of (Abstract, abc.ABC) to preemptively prevent common mro conflicts.

    If not mixed-in with an ABCMeta, it will update __abstractmethods__ itself.
    """

    __slots__ = ()

    __abstractmethods__: ta.ClassVar[ta.FrozenSet[str]] = frozenset()

    #

    def __forceabstract__(self):
        raise TypeError

    # This is done manually, rather than through @abc.abstractmethod, to mask it from static analysis.
    setattr(__forceabstract__, _IS_ABSTRACT_METHOD_ATTR, True)

    #

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        setattr(
            cls,
            _FORCE_ABSTRACT_ATTR,
            getattr(Abstract, _FORCE_ABSTRACT_ATTR) if Abstract in cls.__bases__ else False,
        )

        super().__init_subclass__(**kwargs)

        if not (Abstract in cls.__bases__ or abc.ABC in cls.__bases__):
            if ams := compute_abstract_methods(cls):
                amd = {
                    a: mcls
                    for mcls in cls.__mro__[::-1]
                    for a in ams
                    if a in mcls.__dict__
                }

                raise AbstractTypeError(
                    f'Cannot subclass abstract class {cls.__name__} with abstract methods: ' +
                    ', '.join(sorted([
                        '.'.join([
                            *([
                                *([m] if (m := getattr(c, '__module__')) else []),
                                getattr(c, '__qualname__', getattr(c, '__name__')),
                            ] if c is not None else '?'),
                            a,
                        ])
                        for a in ams
                        for c in [amd.get(a)]
                    ])),
                )

        xbi = (Abstract, abc.ABC)  # , ta.Generic ?
        bis = [(cls.__bases__.index(b), b) for b in xbi if b in cls.__bases__]
        if bis != sorted(bis):
            raise TypeError(
                f'Abstract subclass {cls.__name__} must have proper base class order of '
                f'({", ".join(getattr(b, "__name__") for b in xbi)}), got: '
                f'({", ".join(getattr(b, "__name__") for _, b in sorted(bis))})',
            )

        if not isinstance(cls, abc.ABCMeta):
            update_abstracts(cls, force=True)



# ruff: noqa: UP007
import typing as ta

from ... import dataclasses as dc
from ... import lang


##


@dc.dataclass(frozen=True)
class SpecialValue(lang.Sealed, lang.Abstract):
    pass


@dc.dataclass(frozen=True)
class Now(SpecialValue, lang.Singleton):
    pass


##


SimpleValue: ta.TypeAlias = ta.Union[
    bool,
    int,
    float,
    str,

    SpecialValue,
]



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import dataclasses as dc
import typing as ta

from ...io.types import BytesLike
from ...lite.abstract import Abstract
from ...lite.check import check
from ..headers import HttpHeaders
from ..parsing import ParsedHttpMessage
from ..versions import HttpVersion


##


class IoPipelineHttpMessageObject(Abstract):
    pass


#


class IoPipelineHttpMessageHead(IoPipelineHttpMessageObject, Abstract):
    @property
    @abc.abstractmethod
    def headers(self) -> HttpHeaders:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def parsed(self) -> ta.Optional[ParsedHttpMessage]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def version(self) -> HttpVersion:
        raise NotImplementedError


#


class FullIoPipelineHttpMessage(IoPipelineHttpMessageObject, Abstract):
    @property
    @abc.abstractmethod
    def head(self) -> IoPipelineHttpMessageHead:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def body(self) -> BytesLike:
        raise NotImplementedError


#


@dc.dataclass(frozen=True)
class IoPipelineHttpMessageChunk(IoPipelineHttpMessageObject, Abstract):
    size: int
    # ext: HttpHeaders

    def __post_init__(self) -> None:
        check.arg(self.size > 0)


#


@dc.dataclass(frozen=True)
class IoPipelineHttpMessageEndChunk(IoPipelineHttpMessageObject, Abstract):
    pass


#


@dc.dataclass(frozen=True)
class IoPipelineHttpMessageLastChunk(IoPipelineHttpMessageObject, Abstract):
    # ext: HttpHeaders
    pass


#


@dc.dataclass(frozen=True)
class IoPipelineHttpMessageChunkedTrailers(IoPipelineHttpMessageObject, Abstract):
    # trailers: HttpHeaders
    # parsed_trailers: ta.Optional[ParsedHttpMessage] = None
    pass


#


@dc.dataclass(frozen=True)
class IoPipelineHttpMessageBodyData(IoPipelineHttpMessageObject, Abstract):
    data: BytesLike

    def __post_init__(self) -> None:
        check.arg(len(self.data) > 0)


#


@dc.dataclass(frozen=True)
class IoPipelineHttpMessageEnd(IoPipelineHttpMessageObject, Abstract):
    pass


#


@dc.dataclass(frozen=True)
class IoPipelineHttpMessageAborted(IoPipelineHttpMessageObject, Abstract):
    reason: ta.Union[str, BaseException]

    @property
    def reason_str(self) -> str:
        if isinstance(r := self.reason, str):
            return r
        elif isinstance(r, BaseException):
            return repr(r)
        else:
            raise TypeError(r)


##


def _un_abstract_pipeline_http_object_classes() -> None:
    # So this is regrettable, but I think the benefits of having the base objects be actual dataclasses outweighs the
    # gnarliness here.
    for cls in [IoPipelineHttpMessageHead, FullIoPipelineHttpMessage]:
        atts = {a for a in cls.__dict__ if not a.startswith('_')}
        for att in atts:
            delattr(cls, att)
        ams = check.isinstance(getattr(cls, '__abstractmethods__'), frozenset)
        setattr(cls, '__abstractmethods__', ams - atts)


_un_abstract_pipeline_http_object_classes()


##


class IoPipelineHttpMessageObjects(Abstract):
    @property
    @abc.abstractmethod
    def _head_type(self) -> ta.Type[IoPipelineHttpMessageHead]:
        raise NotImplementedError

    @abc.abstractmethod
    def _make_head(self, parsed: ParsedHttpMessage) -> IoPipelineHttpMessageHead:
        raise NotImplementedError

    #

    @property
    @abc.abstractmethod
    def _full_type(self) -> ta.Type[FullIoPipelineHttpMessage]:
        raise NotImplementedError

    @abc.abstractmethod
    def _make_full(self, head: IoPipelineHttpMessageHead, body: BytesLike) -> FullIoPipelineHttpMessage:
        raise NotImplementedError

    #

    @property
    @abc.abstractmethod
    def _chunk_type(self) -> ta.Type[IoPipelineHttpMessageChunk]:
        raise NotImplementedError

    @abc.abstractmethod
    def _make_chunk(self, size: int) -> IoPipelineHttpMessageChunk:
        raise NotImplementedError

    #

    @property
    @abc.abstractmethod
    def _end_chunk_type(self) -> ta.Type[IoPipelineHttpMessageEndChunk]:
        raise NotImplementedError

    @abc.abstractmethod
    def _make_end_chunk(self) -> IoPipelineHttpMessageEndChunk:
        raise NotImplementedError

    #

    @property
    @abc.abstractmethod
    def _last_chunk_type(self) -> ta.Type[IoPipelineHttpMessageLastChunk]:
        raise NotImplementedError

    @abc.abstractmethod
    def _make_last_chunk(self) -> IoPipelineHttpMessageLastChunk:
        raise NotImplementedError

    #

    @property
    @abc.abstractmethod
    def _chunked_trailers_type(self) -> ta.Type[IoPipelineHttpMessageChunkedTrailers]:
        raise NotImplementedError

    @abc.abstractmethod
    def _make_chunked_trailers(self) -> IoPipelineHttpMessageChunkedTrailers:
        raise NotImplementedError

    #

    @property
    @abc.abstractmethod
    def _body_data_type(self) -> ta.Type[IoPipelineHttpMessageBodyData]:
        raise NotImplementedError

    @abc.abstractmethod
    def _make_body_data(self, data: BytesLike) -> IoPipelineHttpMessageBodyData:
        raise NotImplementedError

    #

    @property
    @abc.abstractmethod
    def _end_type(self) -> ta.Type[IoPipelineHttpMessageEnd]:
        raise NotImplementedError

    @abc.abstractmethod
    def _make_end(self) -> ta.Any:
        raise NotImplementedError

    #

    @property
    @abc.abstractmethod
    def _aborted_type(self) -> ta.Type[IoPipelineHttpMessageAborted]:
        raise NotImplementedError

    @abc.abstractmethod
    def _make_aborted(self, reason: ta.Union[str, BaseException]) -> IoPipelineHttpMessageAborted:
        raise NotImplementedError



import abc
import typing as ta

from ...lite.abstract import _ABSTRACT_METHODS_ATTR  # noqa
from ...lite.abstract import _FORCE_ABSTRACT_ATTR  # noqa
from ...lite.abstract import _IS_ABSTRACT_METHOD_ATTR  # noqa
from ...lite.abstract import Abstract
from ...lite.abstract import is_abstract_method


T = ta.TypeVar('T')


##


def is_abstract_class(obj: ta.Any) -> bool:
    if bool(getattr(obj, _ABSTRACT_METHODS_ATTR, [])):
        return True

    if isinstance(obj, type):
        if Abstract in obj.__bases__:
            return True

        if (
                Abstract in obj.__mro__
                and getattr(obj.__dict__.get(_FORCE_ABSTRACT_ATTR, None), _IS_ABSTRACT_METHOD_ATTR, False)
        ):
            return True

    return False


def is_abstract(obj: ta.Any) -> bool:
    return is_abstract_method(obj) or is_abstract_class(obj)


##


_INTERNAL_ABSTRACT_ATTRS = frozenset([_FORCE_ABSTRACT_ATTR])


def get_abstracts(cls: type, *, include_internal: bool = False) -> frozenset[str]:
    ms = frozenset(getattr(cls, _ABSTRACT_METHODS_ATTR))
    if not include_internal:
        ms -= _INTERNAL_ABSTRACT_ATTRS
    return ms


##


def make_abstract(obj: T) -> T:
    if callable(obj):
        return abc.abstractmethod(obj)

    elif isinstance(obj, property):
        return ta.cast(T, property(
            abc.abstractmethod(obj.fget) if obj.fget is not None else None,
            abc.abstractmethod(obj.fset) if obj.fset is not None else None,
            abc.abstractmethod(obj.fdel) if obj.fdel is not None else None,
        ))

    elif isinstance(obj, (classmethod, staticmethod)):
        return ta.cast(T, type(obj)(abc.abstractmethod(obj.__func__)))

    else:
        return obj



# @omlish-lite
# @omlish-amalg ../../../omdev/scripts/lib/io/pipelines.py
from .bytes.buffering import InboundBytesBufferingIoPipelineHandler  # noqa
from .bytes.decoders import DelimiterFrameDecoderIoPipelineHandler  # noqa
from .bytes.queues import InboundBytesBufferingQueueIoPipelineHandler  # noqa
from .core import IoPipelineHandler  # noqa
from .drivers.asyncio import AsyncioStreamIoPipelineDriver  # noqa
from .handlers.flatmap import FlatMapIoPipelineHandler  # noqa
from .handlers.fns import FnIoPipelineHandler  # noqa
from .handlers.queues import QueueIoPipelineHandler  # noqa
# from .http.client.requests import PipelineHttpRequestEncoder  # noqa
# from .http.client.responses import PipelineHttpResponseDecoder  # noqa
# from .http.server.requests import PipelineHttpRequestDecoder  # noqa
# from .http.server.responses import PipelineHttpResponseEncoder  # noqa


##



# ruff: noqa: UP045
# @omlish-lite
import threading
import typing as ta

from ..semaphores import AsyncliteSemaphore
from ..semaphores import AsyncliteSemaphores
from .base import SyncAsyncliteApi
from .base import SyncAsyncliteObject


##


class SyncAsyncliteSemaphore(AsyncliteSemaphore, SyncAsyncliteObject):
    def __init__(self, u: threading.Semaphore) -> None:
        super().__init__()

        self._u = u

    async def acquire(self, *, timeout: ta.Optional[float] = None) -> None:
        if not self._u.acquire(blocking=True, timeout=timeout):
            raise TimeoutError

    def acquire_nowait(self) -> bool:
        return self._u.acquire(blocking=False)

    def release(self) -> None:
        self._u.release()


class SyncAsyncliteSemaphores(AsyncliteSemaphores, SyncAsyncliteApi):
    def make_semaphore(self, value: int = 1) -> AsyncliteSemaphore:
        return SyncAsyncliteSemaphore(threading.Semaphore(value))



"""
Features:
 - get (line, offset)
 - invalidate
  - TODO: filewatcher
 - linecache interop
  - tokenize.open - detect encoding - only for .py
 - TODO: directory / path_parts tree?

TODO:
 - read raw, decode (detect)
  - ! index from byteofs -> charofs
   - charofs -> lineno
 - 1-based?
 - maximums?
  - lru? max_size/max_entries?
   - collections.cache?

Notes:
 - linecache is 1-based
 - linecache appends final newline if missing:
    if lines and not lines[-1].endswith('\n'):
        lines[-1] += '\n'
"""
import dataclasses as dc
import io
import os.path
import stat as stat_
import tokenize
import typing as ta

from .. import cached
from .. import lang
from ..os.paths import abs_real_path


##


class TextFileCache:
    class FileStat(ta.NamedTuple):
        size: int
        mtime: float

    class Entry:
        def __init__(
                self,
                cache: 'TextFileCache',
                path: str,
                stat: 'TextFileCache.FileStat',
                *,
                encoding: str | None = None,
        ) -> None:
            super().__init__()

            self._cache = cache
            self._path = path
            self._stat = stat
            self._given_encoding = encoding

        def __repr__(self) -> str:
            return lang.attr_repr(self, 'path', 'stat')

        @property
        def path(self) -> str:
            return self._path

        @property
        def stat(self) -> 'TextFileCache.FileStat':
            return self._stat

        @cached.function
        def raw(self) -> bytes:
            return self._cache._read_file(self._path)  # noqa

        @cached.function
        def encoding(self) -> str:
            if (ge := self._given_encoding) is not None:
                return ge
            return self._cache._determine_encoding(self._path, self.raw())  # noqa

        @cached.function
        def text(self) -> str:
            return self.raw().decode(self.encoding())

        @cached.function
        def lines(self) -> ta.Sequence[str]:
            return self.text().splitlines(keepends=True)

        def safe_line(self, n: int, default: str = '') -> str:
            lines = self.lines()
            if 0 <= n < len(lines):
                return lines[n]
            return default

    #

    @dc.dataclass(frozen=True)
    class Config:
        max_file_size: int | None = 5 * 1024 * 1024

        default_encoding: str = 'utf-8'

    def __init__(
            self,
            config: Config = Config(),
            *,
            locking: lang.DefaultLockable = None,
    ) -> None:
        super().__init__()

        self._config = config

        self._dct: dict[str, TextFileCache.Entry] = {}
        self._lock = lang.default_lock(locking)()

    #

    class Error(Exception):
        pass

    @dc.dataclass()
    class PathError(Error):
        path: str

    class PathTypeError(PathError):
        pass

    class FileTooBigError(PathError):
        pass

    #

    def normalize_path(self, path: str) -> str:
        return abs_real_path(path)

    #

    def _try_stat_file(self, path: str) -> FileStat | Exception:
        try:
            st = os.stat(path)
        except FileNotFoundError as e:
            return e

        if not stat_.S_ISREG(st.st_mode):
            return TextFileCache.PathTypeError(path)

        return TextFileCache.FileStat(
            st.st_size,
            st.st_mtime,
        )

    def _read_file(self, path: str) -> bytes:
        with open(path, 'rb') as f:
            return f.read()

    def _determine_encoding(self, path: str, raw: bytes) -> str:
        if path.endswith('.py'):
            buf = io.BytesIO(raw)
            encoding, _ = tokenize.detect_encoding(buf.readline)
            return encoding

        return self._config.default_encoding

    #

    def _get_entry(
            self,
            path: str,
            *,
            check_stat: bool = False,
            or_raise: bool = False,
    ) -> Entry:
        st_: TextFileCache.FileStat | Exception | None = None
        try:
            e = self._dct[path]

        except KeyError:
            if or_raise:
                raise

        else:
            if (
                    not check_stat or
                    (st_ := self._try_stat_file(path)) == e.stat
            ):
                return e

            del self._dct[path]

        st: TextFileCache.FileStat | Exception
        if st_ is not None:
            st = st_
        else:
            st = self._try_stat_file(path)

        if isinstance(st, Exception):
            raise st

        if (mfs := self._config.max_file_size) is not None and st.size > mfs:
            raise TextFileCache.FileTooBigError(path)

        e = TextFileCache.Entry(
            self,
            path,
            st,
        )

        self._dct[path] = e

        return e

    def get_entry(
            self,
            path: str,
            *,
            check_stat: bool = False,
            or_raise: bool = False,
    ) -> Entry:
        path = self.normalize_path(path)

        with self._lock:
            return self._get_entry(
                path,
                check_stat=check_stat,
                or_raise=or_raise,
            )

    #

    def invalidate(self, path: str) -> bool:
        path = self.normalize_path(path)

        with self._lock:
            return self._dct.pop(path, None) is not None



import io
import typing as ta

from ... import collections as col
from ... import dataclasses as dc
from .dtypes import Datetime
from .dtypes import Integer
from .dtypes import String
from .elements import Column
from .elements import Index
from .elements import PrimaryKey
from .elements import UpdatedAtTrigger
from .tabledefs import TableDef
from .values import Now


##


CREATE_UPDATED_AT_TRIGGER_SRC = """\
create trigger {preamble} {trigger_name}
after update on {table_name}
for each row
when new.{column_name} = old.{column_name}
  begin
  update {table_name}
  set {column_name} = current_timestamp
  where rowid = new.rowid;
end\
"""


CREATE_INDEX_SRC = """\
create index {preamble} {index_name} on {table_name} ({columns})
"""


@dc.dataclass()
class RenderColumn:
    name: str
    type: str
    not_null: bool = False
    primary_key: bool = False
    default: str | None = None


def render_create_statements(
        tbl: TableDef,
        *,
        if_not_exists: bool = False,
) -> list[str]:
    cols: ta.Mapping[str, Column] = col.make_map_by(
        lambda c: c.name,
        tbl.elements.by_type[Column],
        strict=True,
    )

    #

    r_cols: dict[str, RenderColumn] = {}

    for c in cols.values():
        ct: str
        if isinstance(c.type, String):
            ct = 'string'
        elif isinstance(c.type, Integer):
            ct = 'integer'
        elif isinstance(c.type, Datetime):
            ct = 'datetime'
        else:
            raise TypeError(c.type)

        dfl: str | None = None
        if c.default.present:
            dv = c.default.must()

            if isinstance(dv, Now):
                dfl = 'current_timestamp'
            else:
                raise TypeError(dv)

        r_cols[c.name] = RenderColumn(
            c.name,
            ct,
            not_null=not c.nullable,
            default=dfl,
        )

    #

    indexes: list[str] = []

    triggers: list[str] = []

    for e in tbl.elements:
        if isinstance(e, Column):
            pass  # Already handled

        elif isinstance(e, PrimaryKey):
            for pc in e.columns:
                rc = r_cols[pc]
                rc.primary_key = True

        elif isinstance(e, UpdatedAtTrigger):
            triggers.append(CREATE_UPDATED_AT_TRIGGER_SRC.format(
                preamble='if not exists' if if_not_exists else '',
                trigger_name=f'{tbl.name}__trigger__updated_at__{e.column}',
                table_name=tbl.name,
                column_name=e.column,
            ))

        elif isinstance(e, Index):
            if (idx_name := e.name) is None:
                idx_name = '__'.join([
                    tbl.name,
                    'index',
                    *e.columns,
                ])

            indexes.append(CREATE_INDEX_SRC.format(
                preamble='if not exists' if if_not_exists else '',
                index_name=idx_name,
                table_name=tbl.name,
                columns=', '.join(e.columns),
            ))

        else:
            raise TypeError(e)

    #

    cts = io.StringIO()

    cts.write(f'create table')
    if if_not_exists:
        cts.write(' if not exists')
    cts.write(f' {tbl.name} (\n')

    for i, rc in enumerate(r_cols.values()):
        cts.write(f'  {rc.name} {rc.type}')

        if rc.not_null:
            cts.write(' not null')

        if rc.primary_key:
            cts.write(' primary key')

        if rc.default is not None:
            cts.write(f' default {rc.default}')

        if i < len(r_cols) - 1:
            cts.write(',')

        cts.write('\n')

    cts.write(')')

    #

    stmts: list[str] = [cts.getvalue()]

    stmts.extend(indexes)
    stmts.extend(triggers)

    return stmts



# ruff: noqa: UP045
# @omlish-lite
import asyncio
import typing as ta

from ...sync.buffers import SyncBufferRelay
from ..sync import AsyncSyncBufferRelay


T = ta.TypeVar('T')


##


@ta.final
class AsyncioSyncBufferRelay(AsyncSyncBufferRelay[T]):
    def __init__(
            self,
            *,
            event: ta.Optional[asyncio.Event] = None,
            loop: ta.Optional[ta.Any] = None,
    ) -> None:
        if event is None:
            event = asyncio.Event()
        self._event = event
        if loop is None:
            loop = asyncio.get_running_loop()
        self._loop = loop

        self._relay: SyncBufferRelay[T] = SyncBufferRelay(
            wake_fn=lambda: loop.call_soon_threadsafe(event.set),  # type: ignore[arg-type]
        )

    def push(self, *vs: T) -> None:
        self._relay.push(*vs)

    def swap(self) -> ta.Sequence[T]:
        return self._relay.swap()

    async def wait(self) -> None:
        await self._event.wait()
        self._event.clear()



# ruff: noqa: UP006 UP045
# @omlish-lite
import typing as ta

from ...parsing import HttpParser
from ..aggregators import IoPipelineHttpObjectAggregatorDecoder
from ..compression.decompressors import IoPipelineHttpObjectDecompressor
from ..decoders import IoPipelineHttpObjectDecoder
from ..responses import IoPipelineHttpResponseObjects


##


class IoPipelineHttpResponseDecoder(IoPipelineHttpResponseObjects, IoPipelineHttpObjectDecoder):
    _parse_mode: ta.Final = HttpParser.Mode.RESPONSE
    _if_content_length_missing: ta.Final = 'eof'


##


class IoPipelineHttpResponseAggregatorDecoder(
    IoPipelineHttpResponseObjects,
    IoPipelineHttpObjectAggregatorDecoder,
):
    _if_content_length_missing: ta.Final = 'eof'


##


class IoPipelineHttpResponseDecompressor(IoPipelineHttpResponseObjects, IoPipelineHttpObjectDecompressor):
    pass



import contextlib
import types
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from .base import AsyncLifecycle
from .base import Lifecycle
from .controller import AsyncLifecycleController
from .controller import LifecycleController
from .states import LifecycleState
from .states import LifecycleStates
from .unwrap import unwrap_async_lifecycle
from .unwrap import unwrap_lifecycle


T = ta.TypeVar('T')

ContextManagerT = ta.TypeVar('ContextManagerT', bound=ta.ContextManager)
AsyncContextManagerT = ta.TypeVar('AsyncContextManagerT', bound=ta.AsyncContextManager)

LifecycleT = ta.TypeVar('LifecycleT', bound=Lifecycle)
AsyncLifecycleT = ta.TypeVar('AsyncLifecycleT', bound=AsyncLifecycle)


##


@ta.final
@dc.dataclass(frozen=True)
class ContextManagerLifecycle(Lifecycle, lang.Final, ta.Generic[ContextManagerT]):
    cm: ContextManagerT

    @ta.override
    def lifecycle_start(self) -> None:
        self.cm.__enter__()

    @ta.override
    def lifecycle_stop(self) -> None:
        self.cm.__exit__(None, None, None)


@ta.final
@dc.dataclass(frozen=True)
class AsyncContextManagerLifecycle(AsyncLifecycle, lang.Final, ta.Generic[AsyncContextManagerT]):
    cm: AsyncContextManagerT

    @ta.override
    async def lifecycle_start(self) -> None:
        await self.cm.__aenter__()

    @ta.override
    async def lifecycle_stop(self) -> None:
        await self.cm.__aexit__(None, None, None)


##


@ta.final
class LifecycleContextManager(lang.Final, ta.Generic[LifecycleT]):
    def __init__(self, lifecycle: LifecycleT) -> None:
        super().__init__()

        self._lifecycle = lifecycle
        self._controller = lifecycle if isinstance(lifecycle, LifecycleController) else LifecycleController(lifecycle)

    __repr__ = lang.attr_ops(lambda o: (o.lifecycle, o.state)).repr

    #

    @property
    def lifecycle(self) -> LifecycleT:
        return self._lifecycle

    @property
    def controller(self) -> LifecycleController:
        return self._controller

    @property
    def state(self) -> LifecycleState:
        return self._controller.state

    #

    def __enter__(self) -> LifecycleT:
        try:
            self._controller.lifecycle_construct()
            self._controller.lifecycle_start()
        except Exception:
            self._controller.lifecycle_destroy()
            raise
        return self._lifecycle

    def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: types.TracebackType | None,
    ) -> bool | None:
        try:
            if self._controller.state is LifecycleStates.STARTED:
                self._controller.lifecycle_stop()
        except Exception:
            self._controller.lifecycle_destroy()
            raise
        else:
            self._controller.lifecycle_destroy()
        return None


@ta.final
class AsyncLifecycleContextManager(lang.Final, ta.Generic[AsyncLifecycleT]):
    def __init__(self, lifecycle: AsyncLifecycleT) -> None:
        super().__init__()

        self._lifecycle = lifecycle
        self._controller = lifecycle if isinstance(lifecycle, AsyncLifecycleController) else AsyncLifecycleController(lifecycle)  # noqa

    __repr__ = lang.attr_ops(lambda o: (o.lifecycle, o.state)).repr

    #

    @property
    def lifecycle(self) -> AsyncLifecycleT:
        return self._lifecycle

    @property
    def controller(self) -> AsyncLifecycleController:
        return self._controller

    @property
    def state(self) -> LifecycleState:
        return self._controller.state

    #

    async def __aenter__(self) -> AsyncLifecycleT:
        try:
            await self._controller.lifecycle_construct()
            await self._controller.lifecycle_start()
        except Exception:
            await self._controller.lifecycle_destroy()
            raise
        return self._lifecycle

    async def __aexit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: types.TracebackType | None,
    ) -> bool | None:
        try:
            if self._controller.state is LifecycleStates.STARTED:
                await self._controller.lifecycle_stop()
        except Exception:
            await self._controller.lifecycle_destroy()
            raise
        else:
            await self._controller.lifecycle_destroy()
        return None


##


def lifecycle_context_manage(obj: T) -> ta.ContextManager[T]:
    @contextlib.contextmanager
    def inner():
        lc = check.not_none(unwrap_lifecycle(obj))
        with LifecycleContextManager(lc):
            yield obj

    return inner()


def async_lifecycle_context_manage(obj: T) -> ta.AsyncContextManager[T]:
    @contextlib.asynccontextmanager
    async def inner():
        lc = check.not_none(unwrap_async_lifecycle(obj))
        async with AsyncLifecycleContextManager(lc):
            yield obj

    return inner()



# ruff: noqa: UP006 UP007 UP043 UP045
# @omlish-lite
"""
TODO:
 - chunked make_chunk_header - https://datatracker.ietf.org/doc/html/rfc9112#name-chunk-extensions
  - and make_body_data ...
 - fix exception handling lol - do we raise ValueError?? do we return aborted??
"""
import abc
import dataclasses as dc
import typing as ta

from ...io.pipelines.bytes.buffering import InboundBytesBufferingIoPipelineHandler
from ...io.pipelines.bytes.decoders import BytesToMessageDecoderIoPipelineHandler
from ...io.pipelines.core import IoPipelineHandlerContext
from ...io.streams.scanning import ScanningByteStreamBuffer
from ...io.streams.segmented import SegmentedByteStreamBuffer
from ...io.streams.segmented import SegmentedByteStreamBufferView
from ...io.streams.types import MutableByteStreamBuffer
from ...io.streams.utils import ByteStreamBuffers
from ...io.streams.utils import CanByteStreamBuffer
from ...lite.abstract import Abstract
from ...lite.check import check
from ..parsing import HttpParser
from ..parsing import parse_http_message
from .bodymodes import IoPipelineHttpBodyMode
from .bodymodes import IoPipelineHttpBodyModeError
from .objects import IoPipelineHttpMessageHead
from .objects import IoPipelineHttpMessageObjects


##


@dc.dataclass(frozen=True)
class IoPipelineHttpDecodingConfig:
    DEFAULT: ta.ClassVar['IoPipelineHttpDecodingConfig']

    parser_config: ta.Optional[HttpParser.Config] = None

    @dc.dataclass(frozen=True)
    class BufferConfig:
        max_size: ta.Optional[int]
        chunk_size: int

    head_buffer: BufferConfig = BufferConfig(max_size=4 * 1024, chunk_size=4 * 1024)

    max_chunk_size: ta.Optional[int] = None
    chunk_header_buffer: BufferConfig = BufferConfig(max_size=1024, chunk_size=1024)


IoPipelineHttpDecodingConfig.DEFAULT = IoPipelineHttpDecodingConfig()


#


class IoPipelineHttpObjectDecoder(
    IoPipelineHttpMessageObjects,
    InboundBytesBufferingIoPipelineHandler,
    BytesToMessageDecoderIoPipelineHandler,
    Abstract,
):
    def __init__(
            self,
            *,
            config: IoPipelineHttpDecodingConfig = IoPipelineHttpDecodingConfig.DEFAULT,
    ) -> None:
        super().__init__()

        self._config = config

        self._state: IoPipelineHttpObjectDecoder._State = self._HeadState(self)

    #

    def inbound_buffered_bytes(self) -> int:
        if (buf := self._state.buf) is None:
            return 0
        return len(buf)

    #

    @property
    @abc.abstractmethod
    def _parse_mode(self) -> HttpParser.Mode:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _if_content_length_missing(self) -> ta.Literal['empty', 'eof']:
        raise NotImplementedError

    #

    def _decode(
            self,
            ctx: IoPipelineHandlerContext,
            data: CanByteStreamBuffer,
            out: ta.List[ta.Any],
            *,
            final: bool = False,
    ) -> None:
        while True:
            if (ret := self._state.decode(ctx, data, out, final=final)) is None:
                return

            self._state, next_data = ret

            if next_data is None:
                return

            data = next_data

    #

    class _State(Abstract):
        def __init__(self, d: 'IoPipelineHttpObjectDecoder') -> None:
            super().__init__()

            self._d = d

        @property
        def buf(self) -> ta.Optional[MutableByteStreamBuffer]:
            return None

        def _abort(
                self,
                out: ta.List[ta.Any],
                reason: ta.Union[str, BaseException],
                data: ta.Optional[CanByteStreamBuffer] = None,
        ) -> ta.Optional[ta.Tuple['IoPipelineHttpObjectDecoder._State', ta.Optional[CanByteStreamBuffer]]]:
            out.append(self._d._make_aborted(reason))  # noqa
            return (self._d._AbortedState(self._d), data)  # noqa

        @abc.abstractmethod
        def decode(
                self,
                ctx: IoPipelineHandlerContext,
                data: CanByteStreamBuffer,
                out: ta.List[ta.Any],
                *,
                final: bool = False,
        ) -> ta.Optional[ta.Tuple['IoPipelineHttpObjectDecoder._State', ta.Optional[CanByteStreamBuffer]]]:
            raise NotImplementedError

    #

    class _HeadState(_State):
        _buf: ta.Optional[MutableByteStreamBuffer] = None

        @property
        def buf(self) -> ta.Optional[MutableByteStreamBuffer]:
            return self._buf

        def decode(
                self,
                ctx: IoPipelineHandlerContext,
                data: CanByteStreamBuffer,
                out: ta.List[ta.Any],
                *,
                final: bool = False,
        ) -> ta.Optional[ta.Tuple['IoPipelineHttpObjectDecoder._State', ta.Optional[CanByteStreamBuffer]]]:
            if final:
                return self._abort(out, 'EOF before HTTP head complete')

            done = False
            next_mvs: ta.List[memoryview]

            for mv in ByteStreamBuffers.iter_segments(data):
                if done:
                    next_mvs.append(mv)  # noqa
                    continue

                if (buf := self._buf) is None:
                    buf = self._buf = ScanningByteStreamBuffer(SegmentedByteStreamBuffer(
                        max_size=self._d._config.head_buffer.max_size,  # noqa
                        chunk_size=self._d._config.head_buffer.chunk_size,  # noqa
                    ))

                rem_mv: ta.Optional[memoryview] = None

                if (max_buf := buf.max_size) is not None:
                    rem_buf = max_buf - len(buf)

                    if len(mv) > rem_buf:
                        buf.write(mv[:rem_buf])
                        rem_mv = mv[rem_buf:]
                    else:
                        buf.write(mv)

                # Look for end of head
                i = buf.find(b'\r\n\r\n')
                if i < 0:
                    if rem_mv is not None:
                        return self._abort(out, 'Head exceeded max buffer size')

                    continue

                # Extract head
                head_view = buf.split_to(i + 4)

                # Parse and emit head
                raw = head_view.tobytes()
                parsed = parse_http_message(
                    raw,
                    mode=self._d._parse_mode,  # noqa
                    config=self._d._config.parser_config,  # noqa
                )

                head = self._d._make_head(parsed)  # noqa
                out.append(head)

                done = True
                next_mvs = []

                # Forward any remainder bytes (body bytes)
                if len(buf):
                    rem_view = buf.split_to(len(buf))
                    next_mvs.extend(rem_view.segments())

                if rem_mv is not None:
                    next_mvs.append(rem_mv)

            if done:
                return (
                    self._d._BodyModeState(self._d, head),  # noqa
                    SegmentedByteStreamBufferView.or_else(next_mvs, b''),
                )
            else:
                return None

    #

    class _BodyModeState(_State):
        def __init__(self, d: 'IoPipelineHttpObjectDecoder', head: IoPipelineHttpMessageHead) -> None:
            super().__init__(d)

            self._head = head

        def decode(
                self,
                ctx: IoPipelineHandlerContext,
                data: CanByteStreamBuffer,
                out: ta.List[ta.Any],
                *,
                final: bool = False,
        ) -> ta.Optional[ta.Tuple['IoPipelineHttpObjectDecoder._State', ta.Optional[CanByteStreamBuffer]]]:
            try:
                te = IoPipelineHttpBodyMode.select(
                    self._head.headers,
                    if_length_missing=self._d._if_content_length_missing,  # noqa
                )
            except IoPipelineHttpBodyModeError as e:
                return self._abort(out, f'Invalid Transfer-Encoding: {e.reason}')

            if te.mode == 'empty':
                out.append(self._d._make_end())  # noqa
                return (self._d._DoneState(self._d, self._head), data)  # noqa

            elif te.mode == 'eof':
                return (self._d._UntilEofContentState(self._d, self._head), data)  # noqa

            elif te.mode == 'cl':
                return (self._d._ContentLengthContentState(self._d, self._head, check.not_none(te.length)), data)  # noqa

            elif te.mode == 'chunked':
                return (self._d._HeaderChunkedContentState(self._d, self._head), data)  # noqa

            else:
                raise RuntimeError(f'unexpected mode {te!r}')

    #

    class _ContentState(_State, Abstract):
        def __init__(
                self,
                d: 'IoPipelineHttpObjectDecoder',
                head: IoPipelineHttpMessageHead,
        ) -> None:
            super().__init__(d)

            self._head = head

    #

    class _UntilEofContentState(_ContentState):
        def decode(
                self,
                ctx: IoPipelineHandlerContext,
                data: CanByteStreamBuffer,
                out: ta.List[ta.Any],
                *,
                final: bool = False,
        ) -> ta.Optional[ta.Tuple['IoPipelineHttpObjectDecoder._State', ta.Optional[CanByteStreamBuffer]]]:
            for mv in ByteStreamBuffers.iter_segments(data):
                if len(data):
                    out.append(self._d._make_body_data(mv))  # noqa

            if final:
                out.append(self._d._make_end())  # noqa
                return (self._d._DoneState(self._d, self._head), b'')  # noqa
            else:
                return None

    #

    class _ContentLengthContentState(_ContentState):
        def __init__(
                self,
                d: 'IoPipelineHttpObjectDecoder',
                head: IoPipelineHttpMessageHead,
                content_length: int,
        ) -> None:
            check.arg(content_length > 0)

            super().__init__(d, head)

            self._remaining = content_length

        def decode(
                self,
                ctx: IoPipelineHandlerContext,
                data: CanByteStreamBuffer,
                out: ta.List[ta.Any],
                *,
                final: bool = False,
        ) -> ta.Optional[ta.Tuple['IoPipelineHttpObjectDecoder._State', ta.Optional[CanByteStreamBuffer]]]:
            next_mvs: ta.List[memoryview]

            for mv in ByteStreamBuffers.iter_segments(data):
                mvl = len(mv)
                if not mvl:
                    continue

                if self._remaining < 1:
                    next_mvs.append(mv)  # noqa
                    continue

                if self._remaining > mvl:
                    out.append(self._d._make_body_data(mv))  # noqa
                    self._remaining -= mvl

                elif self._remaining == mvl:
                    out.append(self._d._make_body_data(mv))  # noqa
                    out.append(self._d._make_end())  # noqa
                    self._remaining = 0
                    next_mvs = []

                else:
                    out.append(self._d._make_body_data(mv[:self._remaining]))  # noqa
                    out.append(self._d._make_end())  # noqa
                    ofs = self._remaining
                    self._remaining = 0
                    next_mvs = [mv[ofs:]]

            if final and self._remaining > 0:
                return self._abort(out, 'EOF before HTTP body complete')
            elif self._remaining == 0:
                return (
                    self._d._DoneState(self._d, self._head),  # noqa
                    SegmentedByteStreamBufferView.or_else(next_mvs, b''),
                )
            else:
                return None

    #

    class _ChunkedContentState(_ContentState, Abstract):
        pass

    #

    class _HeaderChunkedContentState(_ChunkedContentState):
        _buf: ta.Optional[MutableByteStreamBuffer] = None

        @property
        def buf(self) -> ta.Optional[MutableByteStreamBuffer]:
            return self._buf

        def decode(
                self,
                ctx: IoPipelineHandlerContext,
                data: CanByteStreamBuffer,
                out: ta.List[ta.Any],
                *,
                final: bool = False,
        ) -> ta.Optional[ta.Tuple['IoPipelineHttpObjectDecoder._State', ta.Optional[CanByteStreamBuffer]]]:
            chunk_size: ta.Optional[int] = None
            next_mvs: ta.List[memoryview]

            for mv in ByteStreamBuffers.iter_segments(data):
                if chunk_size is not None:
                    next_mvs.append(mv)  # noqa
                    continue

                if (buf := self._buf) is None:
                    buf = self._buf = ScanningByteStreamBuffer(SegmentedByteStreamBuffer(
                        max_size=self._d._config.chunk_header_buffer.max_size,  # noqa
                        chunk_size=self._d._config.chunk_header_buffer.chunk_size,  # noqa
                    ))

                rem_mv: ta.Optional[memoryview] = None

                if (max_buf := buf.max_size) is not None:
                    rem_buf = max_buf - len(buf)

                    if len(mv) > rem_buf:
                        buf.write(mv[:rem_buf])
                        rem_mv = mv[rem_buf:]
                    else:
                        buf.write(mv)

                # Parse chunk size line: <hex-size>\r\n
                i = buf.find(b'\r\n')
                if i < 0:
                    if rem_mv is not None:
                        return self._abort(out, 'Chunk header exceeded max buffer size')

                    continue

                size_line = buf.split_to(i + 2)

                size_bytes = size_line.tobytes()[:-2]  # Strip \r\n
                try:
                    chunk_size = int(size_bytes, 16)
                except ValueError:
                    return self._abort(out, f'Invalid chunk size: {size_bytes!r}')

                if (mcs := self._d._config.max_chunk_size) is not None and chunk_size > mcs:  # noqa
                    return self._abort(out, f'Content chunk size {chunk_size} exceeds maximum content chunk size: {mcs}')  # noqa

                next_mvs = []

                if len(buf) > 0:
                    next_mvs.extend(buf.segments())

                self._buf = None

                if rem_mv is not None:
                    next_mvs.append(rem_mv)

            if chunk_size is not None:
                if chunk_size == 0:
                    out.append(self._d._make_last_chunk())  # noqa
                    return (
                        self._d._TrailerChunkedContentState(self._d, self._head),  # noqa
                        SegmentedByteStreamBufferView.or_else(next_mvs, b''),
                    )
                else:
                    out.append(self._d._make_chunk(chunk_size))  # noqa
                    return (
                        self._d._DataChunkedContentState(self._d, self._head, chunk_size),  # noqa
                        SegmentedByteStreamBufferView.or_else(next_mvs, b''),
                    )
            elif final:
                return self._abort(out, 'EOF before HTTP chunk header complete')
            else:
                return None

    #

    class _DataChunkedContentState(_ChunkedContentState):
        def __init__(
                self,
                d: 'IoPipelineHttpObjectDecoder',
                head: IoPipelineHttpMessageHead,
                chunk_size: int,
        ) -> None:
            check.arg(chunk_size > 0)

            super().__init__(d, head)

            self._remaining = chunk_size

        _got_cr = False

        def decode(
                self,
                ctx: IoPipelineHandlerContext,
                data: CanByteStreamBuffer,
                out: ta.List[ta.Any],
                *,
                final: bool = False,
        ) -> ta.Optional[ta.Tuple['IoPipelineHttpObjectDecoder._State', ta.Optional[CanByteStreamBuffer]]]:
            next_mvs: ta.Optional[ta.List[memoryview]] = None

            for mv in ByteStreamBuffers.iter_segments(data):
                if next_mvs is not None:
                    next_mvs.append(mv)
                    continue

                mvl = len(mv)
                if mvl < 1:
                    continue

                if mvl < self._remaining:
                    self._remaining -= mvl
                    out.append(self._d._make_body_data(mv))  # noqa
                    continue

                if self._remaining > 0:
                    if mvl == self._remaining:
                        out.append(self._d._make_body_data(mv))  # noqa
                        self._remaining = 0
                        continue

                    out.append(self._d._make_body_data(mv[:self._remaining]))  # noqa
                    mv = mv[self._remaining:]
                    mvl = len(mv)
                    self._remaining = 0

                if mvl < 1:
                    continue

                if not self._got_cr:
                    if mv[0] != 0x0d:
                        return self._abort(out, f'Expected \\r\\n after chunk data, got {bytes([mv[0]])!r}')
                    self._got_cr = True
                    mv = mv[1:]
                    mvl -= 1
                    if mvl < 1:
                        continue

                if mv[0] != 0x0a:
                    return self._abort(out, f'Expected \\r\\n after chunk data, got {bytes([mv[0]])!r}')
                mv = mv[1:]
                mvl -= 1

                next_mvs = []

                if mvl > 0:
                    next_mvs.append(mv)

            if next_mvs is not None:
                out.append(self._d._make_end_chunk())  # noqa
                return (
                    self._d._HeaderChunkedContentState(self._d, self._head),  # noqa
                    SegmentedByteStreamBufferView.or_else(next_mvs, b''),
                )
            elif final:
                return self._abort(out, 'EOF before HTTP chunk complete')
            else:
                return None

    #

    class _TrailerChunkedContentState(_ChunkedContentState):
        _got_cr = False

        def decode(
                self,
                ctx: IoPipelineHandlerContext,
                data: CanByteStreamBuffer,
                out: ta.List[ta.Any],
                *,
                final: bool = False,
        ) -> ta.Optional[ta.Tuple['IoPipelineHttpObjectDecoder._State', ta.Optional[CanByteStreamBuffer]]]:
            next_mvs: ta.Optional[ta.List[memoryview]] = None

            for mv in ByteStreamBuffers.iter_segments(data):
                if next_mvs is not None:
                    next_mvs.append(mv)
                    continue

                mvl = len(mv)
                if mvl < 1:
                    continue

                if not self._got_cr:
                    if mv[0] != 0x0d:
                        return self._abort(out, f'Expected \\r\\n after final chunk, got {bytes([mv[0]])!r}')
                    self._got_cr = True
                    mv = mv[1:]
                    mvl -= 1
                    if mvl < 1:
                        continue

                if mv[0] != 0x0a:
                    return self._abort(out, f'Expected \\r\\n after final chunk, got {bytes([mv[0]])!r}')
                mv = mv[1:]
                mvl -= 1

                out.append(self._d._make_chunked_trailers())  # noqa
                out.append(self._d._make_end())  # noqa

                next_mvs = []

                if mvl > 0:
                    next_mvs.append(mv)

            if next_mvs is not None:
                return (
                    self._d._DoneState(self._d, self._head),  # noqa
                    SegmentedByteStreamBufferView.of_opt(next_mvs),
                )
            elif final:
                return self._abort(out, 'EOF before HTTP trailer complete')
            else:
                return None

    #

    class _DoneState(_State):
        def __init__(
                self,
                d: 'IoPipelineHttpObjectDecoder',
                head: ta.Optional[IoPipelineHttpMessageHead] = None,
        ) -> None:
            super().__init__(d)

            self._head = head

        def decode(
                self,
                ctx: IoPipelineHandlerContext,
                data: CanByteStreamBuffer,
                out: ta.List[ta.Any],
                *,
                final: bool = False,
        ) -> ta.Optional[ta.Tuple['IoPipelineHttpObjectDecoder._State', ta.Optional[CanByteStreamBuffer]]]:
            if not len(data):
                return None

            raise NotImplementedError

    #

    class _AbortedState(_State):
        def decode(
                self,
                ctx: IoPipelineHandlerContext,
                data: CanByteStreamBuffer,
                out: ta.List[ta.Any],
                *,
                final: bool = False,
        ) -> ta.Optional[ta.Tuple['IoPipelineHttpObjectDecoder._State', ta.Optional[CanByteStreamBuffer]]]:
            raise NotImplementedError



# ruff: noqa: UP007 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta

from ...io.types import BytesLike
from ...lite.abstract import Abstract
from ...lite.check import check
from ...lite.dataclasses import install_dataclass_kw_only_init
from ..headers import CanHttpHeaders
from ..headers import HttpHeaders
from ..parsing import ParsedHttpMessage
from ..versions import HttpVersion
from ..versions import HttpVersions
from .objects import FullIoPipelineHttpMessage
from .objects import IoPipelineHttpMessageAborted
from .objects import IoPipelineHttpMessageBodyData
from .objects import IoPipelineHttpMessageChunk
from .objects import IoPipelineHttpMessageChunkedTrailers
from .objects import IoPipelineHttpMessageEnd
from .objects import IoPipelineHttpMessageEndChunk
from .objects import IoPipelineHttpMessageHead
from .objects import IoPipelineHttpMessageLastChunk
from .objects import IoPipelineHttpMessageObject
from .objects import IoPipelineHttpMessageObjects


##


class IoPipelineHttpRequestObject(IoPipelineHttpMessageObject, Abstract):
    pass


#


@ta.final
@install_dataclass_kw_only_init()
@dc.dataclass(frozen=True)
class IoPipelineHttpRequestHead(IoPipelineHttpMessageHead, IoPipelineHttpRequestObject):
    method: str
    target: str

    headers: HttpHeaders
    parsed: ta.Optional[ParsedHttpMessage] = None

    version: HttpVersion = HttpVersions.HTTP_1_1


#


@ta.final
@dc.dataclass(frozen=True)
class FullIoPipelineHttpRequest(FullIoPipelineHttpMessage, IoPipelineHttpRequestObject):
    head: IoPipelineHttpRequestHead
    body: BytesLike

    @classmethod
    def simple(
            cls,
            host: str,
            target: str,
            *,
            method: str = 'GET',
            version: HttpVersion = HttpVersions.HTTP_1_1,

            content_type: ta.Optional[str] = None,
            body: bytes = b'',
            connection: str = 'close',

            headers: ta.Optional[CanHttpHeaders] = None,
    ) -> 'FullIoPipelineHttpRequest':
        headers = HttpHeaders.of(headers).update(
            ('Host', host),
            ('Content-Type', content_type),
            ('Content-Length', lambda: str(len(body)) if body else None),
            ('Connection', connection),
            if_present='skip',
        )

        return cls(
            head=IoPipelineHttpRequestHead(
                method=method,
                target=target,
                version=version,
                headers=headers,
            ),
            body=body,
        )


#


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpRequestChunk(IoPipelineHttpMessageChunk, IoPipelineHttpRequestObject):
    pass


#


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpRequestEndChunk(IoPipelineHttpMessageEndChunk, IoPipelineHttpRequestObject):
    pass


#


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpRequestLastChunk(IoPipelineHttpMessageLastChunk, IoPipelineHttpRequestObject):
    pass


#


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpRequestChunkedTrailers(IoPipelineHttpMessageChunkedTrailers, IoPipelineHttpRequestObject):
    pass


#


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpRequestBodyData(IoPipelineHttpMessageBodyData, IoPipelineHttpRequestObject):
    pass


#


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpRequestEnd(IoPipelineHttpMessageEnd, IoPipelineHttpRequestObject):
    pass


#


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpRequestAborted(IoPipelineHttpMessageAborted, IoPipelineHttpRequestObject):
    pass


##


class IoPipelineHttpRequestObjects(IoPipelineHttpMessageObjects):
    _head_type: ta.Final = IoPipelineHttpRequestHead

    def _make_head(self, parsed: ParsedHttpMessage) -> IoPipelineHttpRequestHead:
        request = check.not_none(parsed.request_line)

        return IoPipelineHttpRequestHead(
            method=request.method,
            target=check.not_none(request.request_target).decode('utf-8'),
            version=request.http_version,
            headers=HttpHeaders(parsed.headers.entries),
            parsed=parsed,
        )

    #

    _full_type: ta.Final = FullIoPipelineHttpRequest

    def _make_full(self, head: IoPipelineHttpMessageHead, body: BytesLike) -> FullIoPipelineHttpRequest:
        return FullIoPipelineHttpRequest(check.isinstance(head, IoPipelineHttpRequestHead), body)

    #

    _chunk_type: ta.Final = IoPipelineHttpRequestChunk

    def _make_chunk(self, size: int) -> IoPipelineHttpRequestChunk:
        return IoPipelineHttpRequestChunk(size)

    #

    _end_chunk_type: ta.Final = IoPipelineHttpRequestEndChunk

    def _make_end_chunk(self) -> IoPipelineHttpRequestEndChunk:
        return IoPipelineHttpRequestEndChunk()

    #

    _last_chunk_type: ta.Final = IoPipelineHttpRequestLastChunk

    def _make_last_chunk(self) -> IoPipelineHttpRequestLastChunk:
        return IoPipelineHttpRequestLastChunk()

    #

    _chunked_trailers_type: ta.Final = IoPipelineHttpRequestChunkedTrailers

    def _make_chunked_trailers(self) -> IoPipelineHttpRequestChunkedTrailers:
        return IoPipelineHttpRequestChunkedTrailers()

    #

    _body_data_type: ta.Final = IoPipelineHttpRequestBodyData

    def _make_body_data(self, data: BytesLike) -> IoPipelineHttpRequestBodyData:
        return IoPipelineHttpRequestBodyData(data)

    #

    _end_type: ta.Final = IoPipelineHttpRequestEnd

    def _make_end(self) -> IoPipelineHttpRequestEnd:
        return IoPipelineHttpRequestEnd()

    #

    _aborted_type: ta.Final = IoPipelineHttpRequestAborted

    def _make_aborted(self, reason: ta.Union[str, BaseException]) -> IoPipelineHttpRequestAborted:
        return IoPipelineHttpRequestAborted(reason)



import typing as ta

from ..core import IoPipelineService
from .types import IoPipelineFlow


##


@ta.final
class StubIoPipelineFlowService(IoPipelineFlow, IoPipelineService):
    def __init__(self, *, auto_read: bool = True) -> None:
        super().__init__()

        self._auto_read = auto_read

    def is_auto_read(self) -> bool:
        return self._auto_read



"""
TODO:
 - proper ManagedProvider[Impl], DeferringProvider[Impl], closing (-> Deferring('close'))
"""
import contextlib
import typing as ta

from ... import lang
from ..binder import bind
from ..elements import Elemental
from ..impl.inspect import build_kwargs_target


if ta.TYPE_CHECKING:
    from .. import injector as _injector
    from .. import maysync as _maysync
    from .. import sync as _sync
else:
    _injector = lang.proxy_import('..injector', __package__)
    _maysync = lang.proxy_import('..maysync', __package__)
    _sync = lang.proxy_import('..sync', __package__)


T = ta.TypeVar('T')


##


def create_async_managed_injector(*args: Elemental) -> ta.AsyncContextManager['_injector.AsyncInjector']:
    @contextlib.asynccontextmanager
    async def inner():
        async with contextlib.AsyncExitStack() as aes:
            yield await _injector.create_async_injector(
                bind(contextlib.AsyncExitStack, to_const=aes),
                *args,
            )
    return inner()


def make_async_managed_provider(
        fac: ta.Callable[..., T],
        *fns: ta.Callable[[T], ta.AsyncContextManager[T]],
) -> ta.Callable[..., ta.Awaitable[T]]:
    kt = build_kwargs_target(fac)

    async def _provide(
            ai: _injector.AsyncInjector,
            aes: contextlib.AsyncExitStack,
    ):
        obj = await ai.inject(kt)
        if not fns:
            obj = await aes.enter_async_context(obj)
        else:
            for fn in fns:
                await aes.enter_async_context(fn(obj))
        return obj

    return _provide


##


def create_managed_injector(*args: Elemental) -> ta.ContextManager['_sync.Injector']:
    @contextlib.contextmanager
    def inner():
        with contextlib.ExitStack() as es:
            yield _sync.create_injector(
                bind(contextlib.ExitStack, to_const=es),
                *args,
            )
    return inner()


def make_managed_provider(
        fac: ta.Callable[..., T],
        *fns: ta.Callable[[T], ta.ContextManager[T]],
) -> ta.Callable[..., T]:
    kt = build_kwargs_target(fac)

    def _provide(
            i: _sync.Injector,
            es: contextlib.ExitStack,
    ):
        obj = i.inject(kt)
        if not fns:
            obj = es.enter_context(obj)
        else:
            for fn in fns:
                es.enter_context(fn(obj))
        return obj

    return _provide


##


def create_maysync_managed_injector(*args: Elemental) -> ta.ContextManager['_maysync.MaysyncInjector']:
    @contextlib.contextmanager
    def inner():
        with contextlib.ExitStack() as es:
            yield _maysync.create_maysync_injector(
                bind(contextlib.ExitStack, to_const=es),
                *args,
            )
    return inner()


def make_maysync_managed_provider(
        fac: ta.Callable[..., T],
        *fns: ta.Callable[[T], ta.ContextManager[T]],
) -> ta.Callable[..., T]:
    kt = build_kwargs_target(fac)

    def _provide(
            i: _maysync.MaysyncInjector,
            es: contextlib.ExitStack,
    ):
        obj = i.inject(kt)
        if not fns:
            obj = es.enter_context(obj)
        else:
            for fn in fns:
                es.enter_context(fn(obj))
        return obj

    return _provide



"""
TODO:
 - move to sql/alchemy/tabledefs.py
"""
import typing as ta

import sqlalchemy as sa
import sqlalchemy.sql.schema

from ..tabledefs import TableDef


##


def build_td_table(
        td: TableDef,
        *,
        metadata: sa.MetaData | None = None,
        **kwargs: ta.Any,
) -> sa.Table:
    items: list[sa.sql.schema.SchemaItem] = []

    return sa.Table(
        td.name,
        metadata if metadata is not None else sa.MetaData(),
        *items,
        **kwargs,
    )



# ruff: noqa: UP006 UP045
# @omlish-lite
import collections.abc
import dataclasses as dc
import enum
import functools
import types
import typing as ta

from .....io.pipelines.asyncs import AsyncIoPipelineMessages
from .....io.pipelines.core import IoPipelineHandler
from .....io.pipelines.core import IoPipelineHandlerContext
from .....io.pipelines.core import IoPipelineMessages
from .....io.pipelines.flow.types import IoPipelineFlow
from .....io.pipelines.flow.types import IoPipelineFlowMessages
from .....lite.abstract import Abstract
from .....lite.check import check
from ....headers import HttpHeaders
from ...requests import FullIoPipelineHttpRequest
from ...responses import IoPipelineHttpResponseHead


T = ta.TypeVar('T')


##


@dc.dataclass(frozen=True)
class AsgiSpec:
    app: ta.Any
    host: str = '127.0.0.1'
    port: int = 8087


##


class _AsgiOp(Abstract):
    pass


@dc.dataclass(frozen=True)
class _ReceiveAsgiOp(_AsgiOp):
    pass


@dc.dataclass(frozen=True)
class _SendAsgiOp(_AsgiOp):
    msg: ta.Any


#


class _AsgiFutureNotAwaitedError(RuntimeError):
    pass


class _AsgiFuture(ta.Generic[T]):
    def __init__(self, arg: ta.Any) -> None:
        self.arg = arg

    done: bool = False
    result: T
    error: ta.Optional[BaseException] = None

    def __await__(self) -> ta.Generator['_AsgiFuture', None, T]:
        if not self.done:
            yield self
        if not self.done:
            raise _AsgiFutureNotAwaitedError
        if self.error is not None:
            raise self.error
        else:
            return self.result


#


class _AsgiPump:
    def __init__(self, fn: ta.Any) -> None:
        super().__init__()

        self.fn = fn

    o: ta.Any
    a: ta.Any
    g: ta.Any

    def start(self) -> None:
        self.o = self.fn(self._receive, self._send)
        self.a = self.o.__await__()
        self.g = iter(self.a)

    def close(self) -> None:
        o = getattr(self, 'o', None)
        a = getattr(self, 'a', None)
        g = getattr(self, 'g', None)
        if g is not None and g is not a:
            g.close()
        if a is not None:
            a.close()
        if o is not None:
            o.close()

    @types.coroutine
    def _receive(self) -> ta.Any:
        return _AsgiFuture(_ReceiveAsgiOp())  # type: ignore

    @types.coroutine
    def _send(self, msg: ta.Any) -> ta.Any:
        return _AsgiFuture(_SendAsgiOp(msg))  # type: ignore


class _AsgiDriver:
    def __init__(self, ctx: IoPipelineHandlerContext, fn: ta.Any) -> None:
        super().__init__()

        self._ctx = ctx
        self._pump = _AsgiPump(fn)

    class State(enum.Enum):
        NEW = 'new'

        STARTING = 'starting'
        STARTED = 'started'

        RESPONSE_STARTED = 'response_started'
        RESPONSE_FINISHED = 'response_finished'

        CLOSING = 'closing'
        CLOSED = 'closed'

    _state: State = State.NEW

    def start(self) -> None:
        self._state = _AsgiDriver.State.NEW
        self._pump.start()
        self._state = _AsgiDriver.State.STARTED

    def close(self) -> None:
        if self._state in (_AsgiDriver.State.CLOSING, _AsgiDriver.State.CLOSED):
            return
        self._state = _AsgiDriver.State.CLOSING
        self._pump.close()
        self._state = _AsgiDriver.State.CLOSED

    #

    class _Gv(ta.NamedTuple):
        k: ta.Literal['y', 'r']
        v: ta.Any

    def step(self) -> None:
        if self._state == _AsgiDriver.State.NEW:
            self.start()
        check.state(self._state not in (
            _AsgiDriver.State.NEW,
            _AsgiDriver.State.STARTING,
            _AsgiDriver.State.CLOSING,
            _AsgiDriver.State.CLOSED,
        ))

        out: ta.List[ta.Any] = []

        while True:
            try:
                y = self._pump.g.send(None)  # noqa
            except StopIteration as si:
                r = si.value  # noqa
                check.not_isinstance(r, _AsgiFuture)
                gv = self._Gv('r', r)
            else:
                gv = self._Gv('y', y)

            if not self._step_one(gv, out):
                break

        for out_msg in out:
            self._ctx.feed_out(out_msg)

    #

    def _step_one(self, gv: _Gv, out: ta.List[ta.Any]) -> bool:
        if gv.k == 'y' and not isinstance(gv.v, _AsgiFuture):
            awm = AsyncIoPipelineMessages.Await(gv.v)
            awm.add_listener(lambda _: self.step())
            out.append(awm)
            return False

        if self._state == _AsgiDriver.State.STARTED:
            check.state(gv.k == 'y')
            f = check.isinstance(gv.v, _AsgiFuture)
            md = check.isinstance(check.isinstance(f.arg, _SendAsgiOp).msg, collections.abc.Mapping)
            check.equal(md['type'], 'http.response.start')

            out.append(IoPipelineHttpResponseHead(
                status=(status_code := md['status']),
                reason=IoPipelineHttpResponseHead.get_reason_phrase(status_code),
                headers=HttpHeaders(md['headers']),
            ))
            if IoPipelineFlow.is_auto_read(self._ctx):
                out.append(IoPipelineFlowMessages.FlushOutput())

            self._state = _AsgiDriver.State.RESPONSE_STARTED

            f.result, f.done = None, True
            return True

        elif self._state == _AsgiDriver.State.RESPONSE_STARTED:
            check.state(gv.k == 'y')
            f = check.isinstance(gv.v, _AsgiFuture)
            md = check.isinstance(check.isinstance(f.arg, _SendAsgiOp).msg, collections.abc.Mapping)
            check.equal(md['type'], 'http.response.body')

            out.append(md['body'])
            if IoPipelineFlow.is_auto_read(self._ctx):
                out.append(IoPipelineFlowMessages.FlushOutput())

            if not md.get('more_body', False):
                self._state = _AsgiDriver.State.RESPONSE_FINISHED

            f.result, f.done = None, True
            return True

        elif self._state == _AsgiDriver.State.RESPONSE_FINISHED:
            check.state(gv.k == 'r')
            check.state(gv.v is None)

            out.append(IoPipelineMessages.FinalOutput())

            self.close()
            return False

        else:
            raise RuntimeError(f'Invalid state: {self._state!r}')


#


class AsgiHandler(IoPipelineHandler):
    def __init__(self, app: ta.Any) -> None:
        super().__init__()

        self._app = app

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if isinstance(msg, IoPipelineMessages.InitialInput):
            ctx.feed_in(msg)

            if not IoPipelineFlow.is_auto_read(ctx):
                ctx.feed_out(IoPipelineFlowMessages.ReadyForInput())

            return

        if not isinstance(msg, FullIoPipelineHttpRequest):
            ctx.feed_in(msg)
            return

        #

        scope = {
            'asgi': {
                'spec_version': '2.3',
                'version': '3.0',
            },
            # 'client': ('127.0.0.1', 57782),
            'headers': [
                (k.encode(), v.encode())
                for k, v in msg.head.headers.all
            ],
            'http_version': '1.1',
            'method': msg.head.method,
            'path': msg.head.target,
            # 'query_string': b'',
            # 'raw_path': b'/ping',
            # 'root_path': '',
            'scheme': 'http',
            # 'server': ('127.0.0.1', 8087),
            'state': {},
            'type': 'http',
        }

        #

        drv = _AsgiDriver(ctx, functools.partial(self._app, scope))

        drv.step()



"""
TODO:
 - numeric lettered lists (even unordered) (with separator - `1)` / `1:` / `1.` / ...)
"""
import typing as ta

from ... import check
from .parts import Blank
from .parts import Block
from .parts import Indent
from .parts import List
from .parts import Part
from .parts import Text
from .parts import blockify
from .utils import all_same


##


class ListBuilder:
    DEFAULT_ALLOW_IMPROPER_CHILDREN: ta.ClassVar[bool | ta.Literal['lists_only']] = False
    DEFAULT_LIST_PREFIXES: ta.ClassVar[ta.Sequence[str]] = ['*', '-']

    def __init__(
            self,
            *,
            list_prefixes: ta.Iterable[str] | None = None,
            allow_improper_children: bool | ta.Literal['lists_only'] | None = None,
    ) -> None:
        super().__init__()

        if list_prefixes is None:
            list_prefixes = self.DEFAULT_LIST_PREFIXES
        self._list_prefixes = set(check.not_isinstance(list_prefixes, str))
        if allow_improper_children is None:
            allow_improper_children = self.DEFAULT_ALLOW_IMPROPER_CHILDREN
        self._allow_improper_children = allow_improper_children

        self._len_sorted_list_prefixes = sorted(self._list_prefixes, key=len, reverse=True)

    #

    def _should_promote_indent_child(self, p: Indent) -> bool:
        ac = self._allow_improper_children
        if isinstance(ac, bool):
            return ac
        elif ac == 'lists_only':
            return isinstance(p.p, List)
        else:
            raise TypeError(ac)

    #

    class _DetectedList(ta.NamedTuple):
        pfx: str
        ofs: int
        len: int

    def _detect_list(self, ps: ta.Sequence[Part], st: int = 0) -> _DetectedList | None:
        if not ps:
            return None

        for lp in self._len_sorted_list_prefixes:
            sp = lp + ' '

            mo = -1
            n = st
            while n < len(ps):
                p = ps[n]

                if isinstance(p, (Blank, Text)):
                    if isinstance(p, Text):
                        if p.s.startswith(sp):
                            if mo < 0:
                                mo = n
                        elif mo >= 0:
                            break

                elif isinstance(p, Indent):
                    if mo >= 0 and p.n < len(sp):
                        if not self._should_promote_indent_child(p):
                            break

                elif isinstance(p, List):
                    if mo >= 0:
                        break

                else:
                    raise TypeError(p)

                n += 1

            if mo >= 0:
                return ListBuilder._DetectedList(lp, mo, n - mo)

        return None

    def _build_list(self, lp: str, ps: ta.Sequence[Part]) -> List:
        sp = lp + ' '

        new: list[list[Part]] = []

        f = check.isinstance(ps[0], Text)
        check.state(f.s.startswith(sp))
        new.append([Text(f.s[len(sp):])])
        del f

        for i in range(1, len(ps)):
            p = ps[i]

            if isinstance(p, Blank):
                new[-1].append(p)

            elif isinstance(p, Text):
                check.state(p.s.startswith(sp))
                new.append([Text(p.s[len(sp):])])

            elif isinstance(p, Indent):
                if p.n < len(sp):
                    check.state(self._should_promote_indent_child(p))
                    p = Indent(len(sp), p.p)

                if p.n == len(sp):
                    new[-1].append(p.p)

                else:
                    raise NotImplementedError

            else:
                raise TypeError(p)

        #

        return List(lp, [blockify(*x) for x in new])

    def build_lists(self, root: Part) -> Part:
        def rec(p: Part) -> Part:  # noqa
            if isinstance(p, Block):
                new = [rec(c) for c in p.ps]
                if not all_same(new, p.ps):
                    return rec(blockify(*new))

                st = 0
                diff = False
                while (dl := self._detect_list(new, st)) is not None:
                    diff = True
                    ln = self._build_list(dl.pfx, new[dl.ofs:dl.ofs + dl.len])
                    new[dl.ofs:dl.ofs + dl.len] = [ln]
                    st = dl.ofs + 1

                if diff:
                    p = blockify(*new)
                return p

            elif isinstance(p, Indent):
                if (n := rec(p.p)) is not p.p:
                    p = Indent(p.n, n)  # type: ignore[arg-type]
                return p

            elif isinstance(p, (Blank, Text, List)):
                return p

            else:
                raise TypeError(p)

        return rec(root)



# @omlish-lite
# ruff: noqa: UP045
import errno
import socket
import typing as ta
import urllib.parse

from ....lite.check import check
from ...coro._buffers import ReadableListBuffer
from ...coro.client.connection import CoroHttpClientConnection
from ...coro.client.response import CoroHttpClientResponse
from ...coro.io import CoroHttpIo
from ...headers import HttpHeaders
from ...urls import unparse_url_request_path
from ..base import HttpClientContext
from ..base import HttpClientError
from ..base import HttpClientRequest
from ..sync import HttpClient
from ..sync import StreamHttpClientResponse


T = ta.TypeVar('T')


##


class CoroHttpClient(HttpClient):
    class _Connection:
        def __init__(self, req: HttpClientRequest) -> None:
            super().__init__()

            self._req = req
            self._ups = urllib.parse.urlparse(req.url)

            self._ssl = self._ups.scheme == 'https'

        _cc: ta.Optional[CoroHttpClientConnection] = None
        _resp: ta.Optional[CoroHttpClientResponse] = None

        _sock: ta.Optional[socket.socket] = None
        _sock_file: ta.Optional[ta.BinaryIO] = None

        _ssl_context: ta.Any = None

        #

        def _create_https_context(self, http_version: int) -> ta.Any:
            # https://github.com/python/cpython/blob/a7160912274003672dc116d918260c0a81551c21/Lib/http/client.py#L809
            import ssl

            # Function also used by urllib.request to be able to set the check_hostname attribute on a context object.
            context = ssl.create_default_context()

            # Send ALPN extension to indicate HTTP/1.1 protocol.
            if http_version == 11:
                context.set_alpn_protocols(['http/1.1'])

            # Enable PHA for TLS 1.3 connections if available.
            if context.post_handshake_auth is not None:
                context.post_handshake_auth = True

            return context

        #

        def setup(self) -> StreamHttpClientResponse:
            check.none(self._sock)
            check.none(self._ssl_context)

            self._cc = cc = CoroHttpClientConnection(
                check.not_none(self._ups.hostname),
                default_port=CoroHttpClientConnection.HTTPS_PORT if self._ssl else CoroHttpClientConnection.HTTP_PORT,
            )

            if self._ssl:
                self._ssl_context = self._create_https_context(self._cc.http_version)

            try:
                self._run_coro(cc.connect())

                self._run_coro(cc.request(
                    self._req.method or 'GET',
                    unparse_url_request_path(self._ups) or '/',
                    self._req.data,
                    {k: check.single(v) for k, v in hh.items()} if (hh := self._req.headers_) is not None else {},
                ))

                self._resp = resp = self._run_coro(cc.get_response())

                return StreamHttpClientResponse(
                    status=resp._state.status,  # noqa
                    headers=HttpHeaders(resp._state.headers.items()),  # noqa
                    request=self._req,
                    underlying=self,
                    _stream=ReadableListBuffer().new_bytes_reader(self),
                    _closer=self.close,
                )

            except Exception:
                self.close()
                raise

        def _run_coro(self, g: ta.Generator[ta.Any, ta.Any, T]) -> T:
            i = None

            while True:
                try:
                    o = g.send(i)
                except StopIteration as e:
                    return e.value

                try:
                    i = self._handle_io(o)
                except OSError as e:
                    raise HttpClientError from e

        def _handle_io(self, o: CoroHttpIo.Io) -> ta.Any:
            if isinstance(o, CoroHttpIo.ConnectIo):
                check.none(self._sock)
                self._sock = socket.create_connection(*o.args, **(o.kwargs or {}))

                if self._ssl_context is not None:
                    self._sock = self._ssl_context.wrap_socket(
                        self._sock,
                        server_hostname=check.not_none(o.server_hostname),
                    )

                # Might fail in OSs that don't implement TCP_NODELAY
                try:
                    self._sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                except OSError as e:
                    if e.errno != errno.ENOPROTOOPT:
                        raise

                self._sock_file = self._sock.makefile('rb')

                return None

            elif isinstance(o, CoroHttpIo.CloseIo):
                check.not_none(self._sock).close()
                return None

            elif isinstance(o, CoroHttpIo.WriteIo):
                check.not_none(self._sock).sendall(o.data)
                return None

            elif isinstance(o, CoroHttpIo.ReadIo):
                if (sz := o.sz) is not None:
                    return check.not_none(self._sock_file).read(sz)
                else:
                    return check.not_none(self._sock_file).read()

            elif isinstance(o, CoroHttpIo.ReadLineIo):
                return check.not_none(self._sock_file).readline(o.sz)

            else:
                raise TypeError(o)

        def read1(self, n: int = -1, /) -> bytes:
            return self._run_coro(check.not_none(self._resp).read(n if n >= 0 else None))

        def close(self) -> None:
            if self._resp is not None:
                self._resp.close()
            if self._sock is not None:
                self._sock.close()

    def _stream_request(self, ctx: HttpClientContext, req: HttpClientRequest) -> StreamHttpClientResponse:
        conn = CoroHttpClient._Connection(req)
        return conn.setup()



"""
TODO:
 - options / kwargs
"""
import base64
import binascii
import typing as ta

from .. import check
from .base import Codec
from .funcs import FnPairEagerCodec
from .standard import STANDARD_CODECS


##


class BytesCodec(Codec):
    pass


def make_bytes_encoding_codec(
        name: str,
        aliases: ta.Sequence[str] | None,
        encode: ta.Callable[[bytes], bytes],
        decode: ta.Callable[[bytes], bytes],
        *,
        append_to: ta.MutableSequence[Codec] | None = None,
) -> BytesCodec:
    codec = BytesCodec(
        name=name,
        aliases=check.not_isinstance(aliases, str),

        input=bytes,
        output=bytes,

        new=lambda: FnPairEagerCodec.of(encode, decode),
    )

    if append_to is not None:
        append_to.append(codec)

    return codec


##


ASCII85 = make_bytes_encoding_codec(
    'ascii85',
    ['a85'],
    base64.a85encode,
    base64.a85decode,
    append_to=STANDARD_CODECS,
)

BASE16 = make_bytes_encoding_codec(
    'base16',
    ['b16'],
    base64.b16encode,
    base64.b16decode,
    append_to=STANDARD_CODECS,
)

BASE32 = make_bytes_encoding_codec(
    'base32',
    ['b32'],
    base64.b32encode,
    base64.b32decode,
    append_to=STANDARD_CODECS,
)

BASE64 = make_bytes_encoding_codec(
    'base64',
    ['b64'],
    base64.b64encode,
    base64.b64decode,
    append_to=STANDARD_CODECS,
)

BASE85 = make_bytes_encoding_codec(
    'base85',
    ['b85'],
    base64.b85encode,
    base64.b85decode,
    append_to=STANDARD_CODECS,
)

BASE32_HEX = make_bytes_encoding_codec(
    'base32-hex',
    ['b32-hex'],
    base64.b32hexencode,
    base64.b32hexdecode,
    append_to=STANDARD_CODECS,
)

BASE64_HEX = make_bytes_encoding_codec(
    'base64-hex',
    ['b64-hex'],
    base64.standard_b64encode,
    base64.standard_b64decode,
    append_to=STANDARD_CODECS,
)

BASE64_URLSAFE = make_bytes_encoding_codec(
    'base64-urlsafe',
    ['b64-urlsafe'],
    base64.urlsafe_b64encode,
    base64.urlsafe_b64decode,
    append_to=STANDARD_CODECS,
)

HEX = make_bytes_encoding_codec(
    'hex',
    [],
    binascii.b2a_hex,
    binascii.a2b_hex,
    append_to=STANDARD_CODECS,
)



import keyword
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang


if ta.TYPE_CHECKING:
    import markupsafe as ms

    _HAS_MARKUPSAFE = True

else:
    ms = lang.proxy_import('markupsafe')

    _HAS_MARKUPSAFE = lang.can_import('markupsafe')


String: ta.TypeAlias = ta.Union[
    str,
    'ms.Markup',
]

Content: ta.TypeAlias = ta.Union[
    list['Content'],
    'Dom',
    String,
    None,
]

ContentT = ta.TypeVar('ContentT', bound=Content)


##


ATTR_NAMES_BY_KWARG: ta.Mapping[str, str] = {
    **{f'{k}_': k for k in keyword.kwlist if k == k.lower()},
}

ATTR_KWARGS_BY_NAME: ta.Mapping[str, str] = {v: k for k, v in ATTR_NAMES_BY_KWARG.items()}


def kwargs_to_attrs(**kwargs: ta.Any) -> dict[str, ta.Any]:
    return {
        ATTR_NAMES_BY_KWARG.get(k, k).replace('_', '-'): v
        for k, v in kwargs.items()
    }


##


@dc.dataclass()
class Dom:
    tag: str
    attrs: dict[str, ta.Any | None] | None = dc.xfield(None, repr_fn=lang.opt_repr)
    body: list[Content] | None = dc.xfield(None, repr_fn=lang.opt_repr)

    def set(self, **kwargs: ta.Any) -> 'Dom':
        if self.attrs is None:
            self.attrs = {}
        self.attrs.update(**kwargs_to_attrs(**kwargs))
        return self

    def unset(self, *keys: str) -> 'Dom':
        if self.attrs is not None:
            for k in keys:
                self.attrs.pop(k, None)
        return self

    def add(self, *contents: Content) -> 'Dom':
        if self.body is None:
            self.body = []
        for c in contents:
            check_content(c)
        self.body.extend(contents)
        return self

    def remove(self, *contents: Content, strict: bool = False) -> 'Dom':
        if self.body is not None:
            i = 0
            while i < len(self.body):
                e = self.body[i]
                if any(c is e for c in contents):
                    del self.body[i]
                elif strict:
                    raise ValueError(f'Content {e} not in body')
                else:
                    i += 1
        return self


##


STRING_TYPES: tuple[type, ...] = (
    str,
    *([ms.Markup] if _HAS_MARKUPSAFE else []),
)

CONTENT_TYPES: tuple[type, ...] = (
    list,
    Dom,
    *STRING_TYPES,
    type(None),
)


def check_content(c: ContentT) -> ContentT:
    if isinstance(c, list):
        for e in c:
            check_content(e)
    else:
        check.isinstance(c, CONTENT_TYPES)
    return c


def iter_content(c: Content) -> ta.Iterator[Dom | String]:
    if isinstance(c, list):
        for e in c:
            yield from iter_content(e)
    elif isinstance(c, (Dom, *STRING_TYPES)):
        yield c  # type: ignore[misc]
    elif c is None:
        pass
    else:
        raise TypeError(c)



# @omlish-lite
# ruff: noqa: UP006 UP007 UP037 UP045
# Copyright (c) 2014, Saurabh Kumar (python-dotenv), 2013, Ted Tieken (django-dotenv-rw), 2013, Jacob Kaplan-Moss
# (django-dotenv)
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
#
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
#   disclaimer in the documentation and/or other materials provided with the distribution.
#
# - Neither the name of django-dotenv nor the names of its contributors may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# https://github.com/theskumar/python-dotenv/tree/4d505f2c9bc3569791e64bca0f2e4300f43df0e0/src/dotenv
import abc
import codecs
import contextlib
import io
import os
import pathlib
import re
import shutil
import tempfile
import typing as ta

from ..lite.abstract import Abstract
from ..logs.protocols import LoggerLike


##


_dotenv_posix_variable_pat: ta.Pattern[str] = re.compile(
    r"""
    \$\{
        (?P<name>[^}:]*)
        (?::-
            (?P<default>[^}]*)
        )?
    }
    """,
    re.VERBOSE,
)


class DotenvAtom(Abstract):
    def __ne__(self, other: object) -> bool:
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        return not result

    @abc.abstractmethod
    def resolve(self, env: ta.Mapping[str, ta.Optional[str]]) -> str: ...


class DotenvLiteral(DotenvAtom):
    def __init__(self, value: str) -> None:
        super().__init__()

        self.value = value

    def __repr__(self) -> str:
        return f'DotenvLiteral(value={self.value})'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.value == other.value

    def __hash__(self) -> int:
        return hash((self.__class__, self.value))

    def resolve(self, env: ta.Mapping[str, ta.Optional[str]]) -> str:
        return self.value


class DotenvVariable(DotenvAtom):
    def __init__(self, name: str, default: ta.Optional[str]) -> None:
        super().__init__()

        self.name = name
        self.default = default

    def __repr__(self) -> str:
        return f'DotenvVariable(name={self.name}, default={self.default})'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (self.name, self.default) == (other.name, other.default)

    def __hash__(self) -> int:
        return hash((self.__class__, self.name, self.default))

    def resolve(self, env: ta.Mapping[str, ta.Optional[str]]) -> str:
        default = self.default if self.default is not None else ''
        result = env.get(self.name, default)
        return result if result is not None else ''


def parse_dotenv_variables(value: str) -> ta.Iterator[DotenvAtom]:
    cursor = 0

    for match in _dotenv_posix_variable_pat.finditer(value):
        (start, end) = match.span()
        name = match['name']
        default = match['default']

        if start > cursor:
            yield DotenvLiteral(value=value[cursor:start])

        yield DotenvVariable(name=name, default=default)
        cursor = end

    length = len(value)
    if cursor < length:
        yield DotenvLiteral(value=value[cursor:length])


##


def _make_dotenv_regex(string: str, extra_flags: int = 0) -> ta.Pattern[str]:
    return re.compile(string, re.UNICODE | extra_flags)


_dotenv_newline_pat = _make_dotenv_regex(r'(\r\n|\n|\r)')
_dotenv_multiline_whitespace_pat = _make_dotenv_regex(r'\s*', extra_flags=re.MULTILINE)
_dotenv_whitespace_pat = _make_dotenv_regex(r'[^\S\r\n]*')
_dotenv_export_pat = _make_dotenv_regex(r'(?:export[^\S\r\n]+)?')
_dotenv_single_quoted_key_pat = _make_dotenv_regex(r"'([^']+)'")
_dotenv_unquoted_key_pat = _make_dotenv_regex(r'([^=\#\s]+)')
_dotenv_equal_sign_pat = _make_dotenv_regex(r'(=[^\S\r\n]*)')
_dotenv_single_quoted_value_pat = _make_dotenv_regex(r"'((?:\\'|[^'])*)'")
_dotenv_double_quoted_value_pat = _make_dotenv_regex(r'"((?:\\"|[^"])*)"')
_dotenv_unquoted_value_pat = _make_dotenv_regex(r'([^\r\n]*)')
_dotenv_comment_pat = _make_dotenv_regex(r'(?:[^\S\r\n]*#[^\r\n]*)?')
_dotenv_end_of_line_pat = _make_dotenv_regex(r'[^\S\r\n]*(?:\r\n|\n|\r|$)')
_dotenv_rest_of_line_pat = _make_dotenv_regex(r'[^\r\n]*(?:\r|\n|\r\n)?')
_dotenv_double_quote_escapes_pat = _make_dotenv_regex(r"\\[\\'\"abfnrtv]")
_dotenv_single_quote_escapes_pat = _make_dotenv_regex(r"\\[\\']")


class DotenvOriginal(ta.NamedTuple):
    string: str
    line: int


class DotenvBinding(ta.NamedTuple):
    key: ta.Optional[str]
    value: ta.Optional[str]
    original: DotenvOriginal
    error: bool


class _DotenvPosition:
    def __init__(self, chars: int, line: int) -> None:
        super().__init__()

        self.chars = chars
        self.line = line

    @classmethod
    def start(cls) -> '_DotenvPosition':
        return cls(chars=0, line=1)

    def set(self, other: '_DotenvPosition') -> None:
        self.chars = other.chars
        self.line = other.line

    def advance(self, string: str) -> None:
        self.chars += len(string)
        self.line += len(re.findall(_dotenv_newline_pat, string))


class DotenvError(Exception):
    pass


class _DotenvReader:
    def __init__(self, stream: ta.IO[str]) -> None:
        super().__init__()

        self.string = stream.read()
        self.position = _DotenvPosition.start()
        self.mark = _DotenvPosition.start()

    def has_next(self) -> bool:
        return self.position.chars < len(self.string)

    def set_mark(self) -> None:
        self.mark.set(self.position)

    def get_marked(self) -> DotenvOriginal:
        return DotenvOriginal(
            string=self.string[self.mark.chars:self.position.chars],
            line=self.mark.line,
        )

    def peek(self, count: int) -> str:
        return self.string[self.position.chars:self.position.chars + count]

    def read(self, count: int) -> str:
        result = self.string[self.position.chars:self.position.chars + count]
        if len(result) < count:
            raise DotenvError('read: End of string')
        self.position.advance(result)
        return result

    def read_regex(self, regex: ta.Pattern[str]) -> ta.Sequence[str]:
        match = regex.match(self.string, self.position.chars)
        if match is None:
            raise DotenvError('read_regex: Pattern not found')
        self.position.advance(self.string[match.start():match.end()])
        return match.groups()


def _decode_dotenv_escapes(regex: ta.Pattern[str], string: str) -> str:
    def decode_match(match: ta.Match[str]) -> str:
        return codecs.decode(match.group(0), 'unicode-escape')

    return regex.sub(decode_match, string)


def _parse_dotenv_key(reader: _DotenvReader) -> ta.Optional[str]:
    char = reader.peek(1)
    if char == '#':
        return None
    elif char == "'":
        (key,) = reader.read_regex(_dotenv_single_quoted_key_pat)
    else:
        (key,) = reader.read_regex(_dotenv_unquoted_key_pat)
    return key


def _parse_dotenv_unquoted_value(reader: _DotenvReader) -> str:
    (part,) = reader.read_regex(_dotenv_unquoted_value_pat)
    return re.sub(r'\s+#.*', '', part).rstrip()


def _parse_dotenv_value(reader: _DotenvReader) -> str:
    char = reader.peek(1)
    if char == "'":
        (value,) = reader.read_regex(_dotenv_single_quoted_value_pat)
        return _decode_dotenv_escapes(_dotenv_single_quote_escapes_pat, value)
    elif char == '"':
        (value,) = reader.read_regex(_dotenv_double_quoted_value_pat)
        return _decode_dotenv_escapes(_dotenv_double_quote_escapes_pat, value)
    elif char in ('', '\n', '\r'):
        return ''
    else:
        return _parse_dotenv_unquoted_value(reader)


def _parse_dotenv_binding(reader: _DotenvReader) -> DotenvBinding:
    reader.set_mark()
    try:
        reader.read_regex(_dotenv_multiline_whitespace_pat)
        if not reader.has_next():
            return DotenvBinding(
                key=None,
                value=None,
                original=reader.get_marked(),
                error=False,
            )
        reader.read_regex(_dotenv_export_pat)
        key = _parse_dotenv_key(reader)
        reader.read_regex(_dotenv_whitespace_pat)
        if reader.peek(1) == '=':
            reader.read_regex(_dotenv_equal_sign_pat)
            value: ta.Optional[str] = _parse_dotenv_value(reader)
        else:
            value = None
        reader.read_regex(_dotenv_comment_pat)
        reader.read_regex(_dotenv_end_of_line_pat)
        return DotenvBinding(
            key=key,
            value=value,
            original=reader.get_marked(),
            error=False,
        )
    except DotenvError:
        reader.read_regex(_dotenv_rest_of_line_pat)
        return DotenvBinding(
            key=None,
            value=None,
            original=reader.get_marked(),
            error=True,
        )


def parse_dotenv_stream(stream: ta.IO[str]) -> ta.Iterator[DotenvBinding]:
    reader = _DotenvReader(stream)
    while reader.has_next():
        yield _parse_dotenv_binding(reader)


##


def _dotenv_with_warn_for_invalid_lines(
        mappings: ta.Iterator[DotenvBinding],
        log: ta.Optional[LoggerLike] = None,
) -> ta.Iterator[DotenvBinding]:
    for mapping in mappings:
        if mapping.error:
            if log is not None:
                log.warning(
                    'dotenv could not parse statement starting at line %s',
                    mapping.original.line,
                )
        yield mapping


StrStrMutableMappingT = ta.TypeVar('StrStrMutableMappingT', bound=ta.MutableMapping[str, str])


class Dotenv:
    def __init__(
        self,
        path: ta.Union[str, 'os.PathLike[str]', None] = None,
        stream: ta.Optional[ta.IO[str]] = None,
        verbose: bool = False,
        encoding: ta.Optional[str] = None,
        interpolate: bool = True,
        override: bool = True,
        env: ta.Optional[ta.Mapping[str, str]] = None,
        log: ta.Optional[LoggerLike] = None,
    ) -> None:
        super().__init__()

        self.path: ta.Union[str, 'os.PathLike[str]', None] = path
        self.stream: ta.Optional[ta.IO[str]] = stream
        self._dict: ta.Optional[ta.Dict[str, ta.Optional[str]]] = None
        self.verbose: bool = verbose
        self.encoding: ta.Optional[str] = encoding
        self.interpolate: bool = interpolate
        self.override: bool = override
        self.env = env or {}
        self.log = log

    @contextlib.contextmanager
    def _get_stream(self) -> ta.Iterator[ta.IO[str]]:
        if self.path and os.path.isfile(self.path):
            with open(self.path, encoding=self.encoding) as stream:
                yield stream
        elif self.stream is not None:
            yield self.stream
        else:
            if self.verbose:
                if self.log is not None:
                    self.log.info(
                        'dotenv could not find configuration file %s.',
                        self.path or '.env',
                    )
            yield io.StringIO('')

    def dict(self) -> ta.Dict[str, ta.Optional[str]]:
        if self._dict:
            return self._dict

        raw_values = self.parse()

        if self.interpolate:
            self._dict = dotenv_resolve_variables(raw_values, override=self.override, env=self.env)
        else:
            self._dict = dict(raw_values)

        return self._dict

    def apply_to(self, dst: StrStrMutableMappingT) -> StrStrMutableMappingT:
        for k, v in self.dict().items():
            if v is not None:
                dst[k] = v
            elif k in dst:
                del dst[k]
        return dst

    def parse(self) -> ta.Iterator[ta.Tuple[str, ta.Optional[str]]]:
        with self._get_stream() as stream:
            for mapping in _dotenv_with_warn_for_invalid_lines(parse_dotenv_stream(stream), self.log):
                if mapping.key is not None:
                    yield mapping.key, mapping.value

    def get(self, key: str) -> ta.Optional[str]:
        data = self.dict()

        if key in data:
            return data[key]

        if self.verbose:
            if self.log is not None:
                self.log.warning('Key %s not found in %s.', key, self.path)

        return None


##


def dotenv_get_key(
    path: ta.Union[str, 'os.PathLike[str]'],
    key_to_get: str,
    *,
    encoding: ta.Optional[str] = 'utf-8',
    log: ta.Optional[LoggerLike] = None,
) -> ta.Optional[str]:
    """
    Get the value of a given key from the given .env.

    Returns `None` if the key isn't found or doesn't have a value.
    """

    return Dotenv(
        path,
        verbose=True,
        encoding=encoding,
        log=log,
    ).get(key_to_get)


@contextlib.contextmanager
def _dotenv_rewrite(
    path: ta.Union[str, 'os.PathLike[str]'],
    encoding: ta.Optional[str],
) -> ta.Iterator[ta.Tuple[ta.IO[str], ta.IO[str]]]:
    pathlib.Path(path).touch()

    with tempfile.NamedTemporaryFile(mode='w', encoding=encoding, delete=False) as dest:
        error = None
        try:
            with open(path, encoding=encoding) as source:
                yield (source, dest)
        except BaseException as err:  # noqa
            error = err

    if error is None:
        shutil.move(dest.name, path)
    else:
        os.unlink(dest.name)
        raise error from None


def dotenv_set_key(
    path: ta.Union[str, 'os.PathLike[str]'],
    key_to_set: str,
    value_to_set: str,
    *,
    quote_mode: str = 'always',
    export: bool = False,
    encoding: ta.Optional[str] = 'utf-8',
    log: ta.Optional[LoggerLike] = None,
) -> ta.Tuple[ta.Optional[bool], str, str]:
    """
    Adds or Updates a key/value to the given .env

    If the .env path given doesn't exist, fails instead of risking creating
    an orphan .env somewhere in the filesystem
    """

    if quote_mode not in ('always', 'auto', 'never'):
        raise ValueError(f'Unknown quote_mode: {quote_mode}')

    quote = (
        quote_mode == 'always'
        or (quote_mode == 'auto' and not value_to_set.isalnum())
    )

    if quote:
        value_out = "'{}'".format(value_to_set.replace("'", "\\'"))
    else:
        value_out = value_to_set
    if export:
        line_out = f'export {key_to_set}={value_out}\n'
    else:
        line_out = f'{key_to_set}={value_out}\n'

    with _dotenv_rewrite(path, encoding=encoding) as (source, dest):
        replaced = False
        missing_newline = False
        for mapping in _dotenv_with_warn_for_invalid_lines(parse_dotenv_stream(source), log):
            if mapping.key == key_to_set:
                dest.write(line_out)
                replaced = True
            else:
                dest.write(mapping.original.string)
                missing_newline = not mapping.original.string.endswith('\n')
        if not replaced:
            if missing_newline:
                dest.write('\n')
            dest.write(line_out)

    return True, key_to_set, value_to_set


def dotenv_unset_key(
    path: ta.Union[str, 'os.PathLike[str]'],
    key_to_unset: str,
    *,
    quote_mode: str = 'always',
    encoding: ta.Optional[str] = 'utf-8',
    log: ta.Optional[LoggerLike] = None,
) -> ta.Tuple[ta.Optional[bool], str]:
    """
    Removes a given key from the given `.env` file.

    If the .env path given doesn't exist, fails.
    If the given key doesn't exist in the .env, fails.
    """

    if not os.path.exists(path):
        if log is not None:
            log.warning("Can't delete from %s - it doesn't exist.", path)
        return None, key_to_unset

    removed = False
    with _dotenv_rewrite(path, encoding=encoding) as (source, dest):
        for mapping in _dotenv_with_warn_for_invalid_lines(parse_dotenv_stream(source), log):
            if mapping.key == key_to_unset:
                removed = True
            else:
                dest.write(mapping.original.string)

    if not removed:
        if log is not None:
            log.warning("Key %s not removed from %s - key doesn't exist.", key_to_unset, path)
        return None, key_to_unset

    return removed, key_to_unset


def dotenv_resolve_variables(
        values: ta.Iterable[ta.Tuple[str, ta.Optional[str]]],
        override: bool,
        env: ta.Mapping[str, str],
) -> ta.Dict[str, ta.Optional[str]]:
    new_values: ta.Dict[str, ta.Optional[str]] = {}

    for (name, value) in values:
        if value is None:
            result = None
        else:
            atoms = parse_dotenv_variables(value)
            aenv: ta.Dict[str, ta.Optional[str]] = {}
            if override:
                aenv.update(env)
                aenv.update(new_values)
            else:
                aenv.update(new_values)
                aenv.update(env)
            result = ''.join(atom.resolve(aenv) for atom in atoms)

        new_values[name] = result

    return new_values


def dotenv_values(
    path: ta.Union[str, 'os.PathLike[str]', None] = None,
    stream: ta.Optional[ta.IO[str]] = None,
    *,
    verbose: bool = False,
    interpolate: bool = True,
    encoding: ta.Optional[str] = 'utf-8',
    env: ta.Optional[ta.Mapping[str, str]] = None,
    log: ta.Optional[LoggerLike] = None,
) -> ta.Dict[str, ta.Optional[str]]:
    """
    Parse a .env file and return its content as a dict.

    The returned dict will have `None` values for keys without values in the .env file.
    For example, `foo=bar` results in `{"foo": "bar"}` whereas `foo` alone results in
    `{"foo": None}`

    Args:
        path: Absolute or relative path to the .env file.
        stream: `StringIO` object with .env content, used if `path` is `None`.
        verbose: Whether to output a warning if the .env file is missing.
        encoding: Encoding to be used to read the file.

    If both `path` and `stream` are `None`, `find_dotenv()` is used to find the
    .env file.
    """

    if path is None and stream is None:
        raise ValueError('must set path or stream')

    return Dotenv(
        path=path,
        stream=stream,
        verbose=verbose,
        interpolate=interpolate,
        override=True,
        encoding=encoding,
        env=env,
        log=log,
    ).dict()
